package ru.dbotthepony.kstarbound.defs

import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.reflect.TypeToken
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.util.AssetPathStack
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kstarbound.json.JsonPath
import java.util.function.Function
import java.util.function.Supplier
import kotlin.properties.Delegates
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty
import kotlin.reflect.javaType

/**
 * Base class for instances which can be mutated by altering underlying JSON
 */
abstract class JsonDriven(val path: String) {
	private val delegates = ArrayList<Property<*>>()
	private val lazies = ArrayList<LazyData<*>>()

	/**
	 * [JsonObject]s which define behavior of properties
	 */
	abstract fun lookupProperty(path: JsonPath, orElse: () -> JsonElement): JsonElement

	fun lookupProperty(key: JsonPath): JsonElement {
		return lookupProperty(key) { JsonNull.INSTANCE }
	}

	fun setProperty(key: JsonPath, value: JsonElement) {
		setProperty0(key, value)
		invalidate()
	}

	protected abstract fun setProperty0(key: JsonPath, value: JsonElement)

	protected open fun invalidate() {
		delegates.forEach { it.invalidate() }
		lazies.forEach { it.invalidate() }
	}

	inner class LazyData<T>(private val initializer: () -> T) : Lazy<T> {
		init {
			lazies.add(this)
		}

		private var _value: Any? = mark

		override val value: T get() {
			var value = _value

			if (value !== mark) {
				return value as T
			}

			value = initializer.invoke()
			_value = value
			return value
		}

		override fun isInitialized(): Boolean {
			return _value !== mark
		}

		fun invalidate() {
			_value = mark
		}
	}

	inner class Property<T>(
		val name: JsonPath,
		val default: Either<Supplier<T>, JsonElement>? = null,
		private var adapter: TypeAdapter<T>? = null,
	) : Delegate<T>, ReadWriteProperty<Any?, T> {
		constructor(name: JsonPath, default: T, adapter: TypeAdapter<T>? = null) : this(name, Either.left(Supplier { default }), adapter)
		constructor(name: JsonPath, default: Supplier<T>, adapter: TypeAdapter<T>? = null) : this(name, Either.left(default), adapter)
		constructor(name: JsonPath, default: JsonElement, adapter: TypeAdapter<T>? = null) : this(name, Either.right(default), adapter)

		init {
			delegates.add(this)
		}

		private var value: Supplier<T> by Delegates.notNull()

		private fun compute(): T {
			val value = lookupProperty(name)

			if (value.isJsonNull) {
				if (default == null) {
					throw NoSuchElementException("No json value present at '$name', and no default value was provided")
				} else if (default.isLeft) {
					return default.left().get()
				} else {
					AssetPathStack(this@JsonDriven.path) {
						return adapter!!.fromJsonTree(default.right())
					}
				}
			} else {
				AssetPathStack(this@JsonDriven.path) {
					return adapter!!.fromJsonTree(value)
				}
			}
		}

		fun invalidate() {
			value = Supplier {
				val new = compute()
				this.value = Supplier { new }
				new
			}
		}

		init {
			invalidate()
		}

		override fun get(): T {
			return value.get()
		}

		override fun accept(t: T) {
			AssetPathStack(this@JsonDriven.path) {
				setProperty0(name, adapter!!.toJsonTree(t))
			}

			value = Supplier { t }
		}

		@OptIn(ExperimentalStdlibApi::class)
		override fun getValue(thisRef: Any?, property: KProperty<*>): T {
			if (adapter == null) {
				adapter = Starbound.gson.getAdapter(TypeToken.get(property.returnType.javaType)) as TypeAdapter<T>
			}

			return value.get()
		}

		@OptIn(ExperimentalStdlibApi::class)
		override fun setValue(thisRef: Any?, property: KProperty<*>, value: T) {
			if (adapter == null) {
				adapter = Starbound.gson.getAdapter(TypeToken.get(property.returnType.javaType)) as TypeAdapter<T>
			}

			return accept(value)
		}
	}

	companion object {
		private val mark = Any()
	}
}
