package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kommons.io.readUUID
import ru.dbotthepony.kommons.io.writeUUID
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kstarbound.GlobalDefaults
import ru.dbotthepony.kstarbound.defs.SpawnTarget
import ru.dbotthepony.kstarbound.defs.WarpAction
import ru.dbotthepony.kstarbound.defs.WarpMode
import ru.dbotthepony.kstarbound.defs.WorldID
import ru.dbotthepony.kstarbound.defs.world.AsteroidsWorldParameters
import ru.dbotthepony.kstarbound.defs.world.SkyParameters
import ru.dbotthepony.kstarbound.defs.world.SkyType
import ru.dbotthepony.kstarbound.io.readVector2d
import ru.dbotthepony.kstarbound.io.writeStruct2d
import ru.dbotthepony.kstarbound.network.syncher.legacyCodec
import ru.dbotthepony.kstarbound.network.syncher.nativeCodec
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.UUID
import kotlin.math.PI
import kotlin.math.absoluteValue

sealed class SystemWorldLocation {
	abstract fun write(stream: DataOutputStream, isLegacy: Boolean)
	abstract suspend fun resolve(system: SystemWorld): Vector2d?
	abstract suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>>
	abstract suspend fun skyParameters(system: SystemWorld): SkyParameters

	protected suspend fun appendParameters(parameters: SkyParameters, system: SystemWorld) {

	}

	object Transit : SystemWorldLocation() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(0)
		}

		override suspend fun resolve(system: SystemWorld): Vector2d? {
			return null
		}

		override suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>> {
			return KOptional()
		}

		override suspend fun skyParameters(system: SystemWorld): SkyParameters {
			return GlobalDefaults.systemWorld.emptySkyParameters
		}
	}

	// orbiting around specific planet
	data class Celestial(val position: UniversePos) : SystemWorldLocation() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(1)
			position.write(stream, isLegacy)
		}

		override suspend fun resolve(system: SystemWorld): Vector2d {
			return system.planetPosition(position)
		}

		override suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>> {
			return KOptional(WarpAction.World(WorldID.Celestial(position)) to WarpMode.BEAM_OR_DEPLOY)
		}

		override suspend fun skyParameters(system: SystemWorld): SkyParameters {
			return SkyParameters.create(position, system.universe)
		}
	}

	// orbiting around celestial body
	data class Orbit(val position: SystemWorld.Orbit) : SystemWorldLocation() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(2)
			position.write(stream, isLegacy)
		}

		override suspend fun resolve(system: SystemWorld): Vector2d {
			return system.orbitPosition(position)
		}

		override suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>> {
			return KOptional()
		}

		override suspend fun skyParameters(system: SystemWorld): SkyParameters {
			if (position.target.isPlanet) {

			}

			return GlobalDefaults.systemWorld.emptySkyParameters
		}
	}

	data class Entity(val uuid: UUID) : SystemWorldLocation() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(3)
			stream.writeUUID(uuid)
		}

		override suspend fun resolve(system: SystemWorld): Vector2d? {
			return system.entities[uuid]?.position
		}

		override suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>> {
			val action = system.entities[uuid]?.data?.warpAction ?: return KOptional()
			return KOptional(action to WarpMode.DEPLOY_ONLY)
		}

		override suspend fun skyParameters(system: SystemWorld): SkyParameters {
			val get = system.entities[uuid] ?: return GlobalDefaults.systemWorld.emptySkyParameters
			return get.data.skyParameters
		}
	}

	data class Position(val position: Vector2d) : SystemWorldLocation() {
		override fun write(stream: DataOutputStream, isLegacy: Boolean) {
			stream.writeByte(4)
			stream.writeStruct2d(position, isLegacy)
		}

		override suspend fun resolve(system: SystemWorld): Vector2d {
			return position
		}

		override suspend fun orbitalAction(system: SystemWorld): KOptional<Pair<WarpAction, WarpMode>> {
			// player can beam to asteroid fields simply by being in proximity to them
			for (child in system.universe.children(system.systemLocation)) {
				if ((system.planetPosition(child).length - position.length).absoluteValue > GlobalDefaults.systemWorld.asteroidBeamDistance) {
					continue
				}

				val params = system.universe.parameters(child) ?: continue
				val visitable = params.visitableParameters ?: continue

				if (visitable is AsteroidsWorldParameters) {
					val targetX = position.toAngle() / (2.0 * PI) * visitable.worldSize.x
					return KOptional(WarpAction.World(WorldID.Celestial(child), SpawnTarget.X(targetX)) to WarpMode.DEPLOY_ONLY)
				}
			}

			return KOptional()
		}

		override suspend fun skyParameters(system: SystemWorld): SkyParameters {
			for (child in system.universe.children(system.systemLocation)) {
				if ((system.planetPosition(child).length - position.length).absoluteValue > GlobalDefaults.systemWorld.asteroidBeamDistance) {
					continue
				}

				val params = system.universe.parameters(child) ?: continue
				val visitable = params.visitableParameters ?: continue

				if (visitable is AsteroidsWorldParameters) {
					return SkyParameters.create(child, system.universe)
				}
			}

			return GlobalDefaults.systemWorld.emptySkyParameters
		}
	}

	companion object {
		val CODEC = nativeCodec(::read, SystemWorldLocation::write)
		val LEGACY_CODEC = legacyCodec(::read, SystemWorldLocation::write)

		fun read(stream: DataInputStream, isLegacy: Boolean): SystemWorldLocation {
			return when (val type = stream.readUnsignedByte()) {
				0 -> Transit
				1 -> Celestial(UniversePos(stream, isLegacy))
				2 -> Orbit(SystemWorld.Orbit(stream, isLegacy))
				3 -> Entity(stream.readUUID())
				4 -> Position(stream.readVector2d(isLegacy))
				else -> throw IllegalStateException("Unknown SystemWorldLocation type $type!")
			}
		}
	}
}
